/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package gsi;

/**
 *
 * @author demel
 */
public class Arc {
    Gsi model;
    String   strFrom,  strTo;
    int        iFrom,    iTo;
    Node    nodeFrom, nodeTo;
    Double  probab;
    String  strDistribution;
    enum Distribution {CONST, EXP, UNIFORM, INTUNIFORM, GAUSS, BETA, TRIANGLE};
    Distribution d;
    Double  p1, p2, p3;
    boolean running = false;
    boolean signalling = false;

    public Arc(java.util.Scanner f, Gsi model) {
        this.model = model;
        strFrom = f.next();
        strTo   = f.next();
        probab  = f.nextDouble();
        strDistribution = f.next();
        d = Distribution.valueOf(strDistribution.toUpperCase());
        switch (d) {
            case    CONST:
            case    EXP:
                p1 = f.nextDouble();
                p2 = p3 = 0.0;
                break;
            case    UNIFORM:
            case    GAUSS:
                p1 = f.nextDouble();
                p2 = f.nextDouble();
                p3 = 0.0;
                break;
            case    INTUNIFORM:
            case    TRIANGLE:
            case    BETA:
                p1 = f.nextDouble();
                p2 = f.nextDouble();
                p3 = f.nextDouble();
                break;
        }
        if (probab <= 0.0) {
            model.die("Non positive probability");
        }
        printArc();
    }

    public String toString() {
        return "Arc: " + strFrom + " --> " + strTo +
            "   P=" + probab +
            "   Distribution: " + strDistribution + "   " +
            p1 + "  " + p2 + "  " + p3 + "\n";
    }

    public void printArc() {
        System.out.print(
            "Arc: " + strFrom + " --> " + strTo +
            "   P=" + probab +
            "   Distribution: " + strDistribution + "   " +
            p1 + "  " + p2 + "  " + p3 + "\n");
    }

    public String fromTo() {
        return strFrom + " --> " + strTo;
    }

    public void start() {
        Double duration, t;
        if (running) {
            model.die("Error --- started already running arc  "+fromTo());
        }
        running = true;
        signalling = false;
        duration = genDuration();
        t = model.time + duration;
        Event e = new Event(t, this);
        model.que.add(e);
        model.log(3, model.time+":    arc "+fromTo()+"  started  (duration="+duration+")");
        model.log(4, "...queued event at "+t);
        model.log(4, "...first item in queue is at "+model.que.peek().time);
    }

    public void stop() {
        running = false;
        signalling = true;
        model.log(3, model.time+":    arc "+fromTo()+"  finished");
        nodeTo.notified(this);
    }

    public void reset() {
        running = false;
        signalling = false;
    }

    public Double genDuration() {
        Double x, r;
        model.logc(5, "... generating "+ strDistribution+
                "  "+p1+"  "+p2+"  "+p3+"   --> ");
        switch (d) {
            case CONST:
                x = p1;
                break;
            case UNIFORM:
                x = model.rng.nextDouble() * (p2-p1) + p1;
                break;
            case INTUNIFORM:
                int  k1,k2,k3;
                k1= (int)Math.round(p1); 
                k2= (int)Math.round(p2); 
                k3= (int)Math.round(p3); 
                x = (double) model.rng .nextInt((k2-k1)/k3 + 1) * k3 + k1;
                break;
            case EXP:
                // p1 = mean ,  hence lambda=(1.0/p1)
                do {
                    r = model.rng.nextDouble();
                } while (r == 0.0);
                x = (-p1 * Math.log(r));
                break;
            case GAUSS:
                do {
                    x = model.rng.nextGaussian() * p2 + p1;
                } while (x <= 0.0);
                break;
            case BETA:
                double min, max, mode, mean, stdev, Z, shA, shB;
                min = p1;
                max = p2;
                mode = p3;
                mean = (min+4.0*mode+max)/6.0;
                stdev = (max-min)/6.0;
                Z = (mean - min)/(max-min);
                shA = Z * ((max-mean)*(mean-min)/(stdev*stdev) - 1.0);
                shB = Z * shA;
                x = rBeta(shA, shB);
                return x * (max-min) + min;
              
            case TRIANGLE:
                // http://en.wikipedia.org/wiki/Triangular_distribution
                Double u, f, a, b, c;
                a=p1; b=p2; c=p3;
                if (a==b) {
                    x = a;
                    break;
                }
                f = (c-a)/(b-a);
                u = model.rng.nextDouble();
                if (u < f) {
                    x = a + Math.sqrt(u*(b-a)*(c-a));
                } else {
                    x = b - Math.sqrt(u*(1.0-u)*(b-a)*(b-c));
                }
                break;
            default:
                x = 0.0;
                break;
        }
        model.log(5, " "+x+"\n");
        return x;
    }

    public double rBeta(double alpha, double beta) {
        double rB, gA, gB;
        gA = rGamma(alpha);
        gB = rGamma(beta);
        rB = gA / (gA+gB);
        return rB;
    }
    
    public double rGamma(double alpha) {
        double rG;
        rG = 1.0;
        for (double i=1.0; i <= alpha; i++) {
            rG *= model.rng.nextDouble();
        }
        return Math.log(rG);
    }
}
