/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package gsi;

    import java.util.*;
    import java.io.*;

/**
 *
 * @author demel
 */
public class Gsi {

    List <Node> nodes = new ArrayList<Node>(100);
    List <Arc>  arcs  = new ArrayList<Arc>(100);
    enum Cmds {model, node, arc, simulate, seed};
    String  modelTxt = "";
    Double  stopAtTime = -1.0;  // stop simulation run at this time
    Long    usedRandSeed;
    String  strStartNode = "";
    Node    startNode;
    Double  delta;                  // time tolerance for EOR nodes
    int     nRuns;                  // the number of simulation runs
    public Double time = 0.0;
    Queue <Event> que = new PriorityQueue<Event>();
    List <Node> postponedEorNodes = new ArrayList<Node>(100);
    Double      postponedEorTime;
    Random rng = new Random();

    int     logLevel = 1;
    boolean ok = true;
    // Node   newNode;

    public Gsi() {
        // this hack is to know the seed even if derived from system time
        usedRandSeed = rng.nextLong();
        rng.setSeed(usedRandSeed);
        
        delta = 0.001;
    }

    boolean  readModel(String inFileName) {
        Scanner f;
        Node n, n2;
        Arc  a;
        try {
            f = new Scanner (new File(inFileName));
            f.useLocale(Locale.US);
            String cmdStr = "";
            String rest, word;

            while (f.hasNext() && ok) {
                cmdStr = f.next();
                if (cmdStr.equalsIgnoreCase("model")) {
                    logc(0, cmdStr + " ");
                    rest = f.nextLine();
                    modelTxt += rest + "\n";
                    log(0, rest);
                } else {
                    break;
                }
            }

            mainReadingCycle:
            while (true) {
                if (cmdStr.equalsIgnoreCase("node")) {
                    n = new Node(f,this);
                    n2 = findNodeByName(n.name);
                    if (n2 == null) {
                        nodes.add(n); 
                    }
                    else {    
                        die("Duplicate node name:  " + n.name);
                    }
                }
                else if (cmdStr.equalsIgnoreCase("arc")) {
                    a = new Arc(f,this);
                    arcs.add(a);
                }
                else if (cmdStr.equalsIgnoreCase("startNode")) {
                    if (f.hasNext()) {
                        strStartNode = f.next();
                        log(0, "Start "+strStartNode);
                    } else {
                        die("Missing name of the start node");
                    }
                }
                else if (cmdStr.equalsIgnoreCase("stopAtTime")) {
                    if (f.hasNextDouble()) {
                        stopAtTime = f.nextDouble();
                        log(0, "StopAtTime "+stopAtTime);
                        if (stopAtTime <= 0.0) {
                            die("Final simulation time must be positive");
                        }
                    } else {
                        die("Missing value of the final simulation time");
                    }
                }
                else if (cmdStr.equalsIgnoreCase("logLevel")) {
                    if (f.hasNextInt()) {
                        logLevel = f.nextInt();
                        log(0, "logLevel "+logLevel);
                        if (logLevel < 0 || logLevel > 5) {
                            die("logLevel out of range (0 .. 5)");
                        }
                    } else {
                        die("Missing value of log level");
                    }
                }
                else if (cmdStr.equalsIgnoreCase("seed")) {
                    if (f.hasNextLong()) {
                        usedRandSeed = f.nextLong();
                        log(0, "Seed "+usedRandSeed);
                        rng.setSeed(usedRandSeed);
                    } else {
                        die("Missing value of random seed");
                    }
                }
                else if (cmdStr.equalsIgnoreCase("delta")) {
                    if (f.hasNextDouble()) {
                        delta = f.nextDouble();
                        log(0, "delta "+stopAtTime);
                        if (delta <= 0.0) {
                            die("Delta must be positive");
                        }
                    } else {
                        die("Missing value of delta");
                    }
                }
                else if (cmdStr.equalsIgnoreCase("simulate")) {
                    if (f.hasNextInt()) {
                        nRuns = f.nextInt();
                        log(0, "Simulate "+nRuns);
                        if (nRuns <= 0) {
                            die("Number of simulation runs must be positive");
                        }
                        break mainReadingCycle;
                    } else {
                        die("Missing number of simulation runs");
                    }
                }
                else if (cmdStr.equalsIgnoreCase("//")) {
                    // do nothing -- ignore line
                }
                else {
                    die("Unrecognized command \""+cmdStr+"\"");
                    break;
                }
                if (!ok) {
                    break;
                }
                f.nextLine();
                if (f.hasNext()) {
                    cmdStr = f.next();
                }
            } 
            f.close();
        }
        catch (IOException e) {
            System.out.println("Soubor " + inFileName + " nenalezen, neotevřen");
            ok = false;
        }
        //catch (Exception e) {
        //    System.out.println("Nějaká chyba při čtení");
        //    ok = false;
        //}
        return ok;
    };

    void     die(String txt) {
            System.out.print(txt+"\n");
            ok = false;
    }

    public void printModel() {
        System.out.println("\nVypis modelu:");
        System.out.println("\n"+modelTxt+"\n");
        for (Node n: nodes) {
            n.printNode();
        }
        //System.out.print(nodes);
        for (Arc a: arcs) {
            a.printArc();
        }
        System.out.println("startNode  "+strStartNode);
        System.out.println("stopAtTime "+stopAtTime);
        System.out.println("seed     "+usedRandSeed);
        System.out.println("simulate "+nRuns);
    }

    Node findNodeByName(String pName) {
        for (Node n: nodes) {
            if (pName.equalsIgnoreCase(n.name))  return n;
        }
        return null;
    }

    public void checkModel() {
        Node n2;
        for (Node n: nodes) {
            n2 = findNodeByName(n.name);
            if (n2 != n) {
                die("Duplicate node name:  xxx" + n.name);
            }
        }

        if (strStartNode.isEmpty()) {
            die("Start node not given");
        } else {
            startNode = findNodeByName(strStartNode);
            if (startNode == null) {
                die("Start node "+strStartNode+" not found");
            }
        }

        if (stopAtTime == -1.0) {
            die("stopAtTime not given. It must be positive]");
        }

        for (Arc arc: arcs) {
            arc.nodeFrom = findNodeByName(arc.strFrom);
            if (arc.nodeFrom==null) {
                arc.printArc();
                die ("Source node not defined:  " + arc.strFrom);
            } else {
                arc.nodeFrom.fwdLinks.add(arc);
            }
            arc.nodeTo   = findNodeByName(arc.strTo);
            if (arc.nodeTo  ==null) {
                arc.printArc();
                die ("Target node not defined:  " + arc.strTo);
            } else {
                arc.nodeTo.bwdLinks.add(arc);
            }
        }               // end of the loop over arcs


        for (Node n: nodes) {
            if (n.typOut == Node.TypOut.ST  &&  n.fwdLinks.size() > 0) {
                Double sumP = 0.0;
                for (Arc a: n.fwdLinks) {
                    sumP += a.probab;
                }
                if (Math.abs(sumP - 1.0) > 0.001) {
                    for (Arc a: n.fwdLinks) {
                        a.probab /= sumP;
                    }
                    log(0, "probabilities out of node "+n.name+" rescaled upto 1.0");
                }
            }
        }

        if (ok) {
            log(0, "----- Model linked and verified -----\n");
        }
        else {
            log(0, "----- Model is invalid. -----\n");
        }
    }

    public void simulate() {
        Event  event;
        //Node n;
        int  run;

        for (run=1; run <= nRuns; run++) {
            // initialize next simulation run
            log(1, "----- started simulaion run "+run+" out of "+nRuns);
            time = 0.0;
            que.clear();
            postponedEorNodes.clear();
            postponedEorTime = -1.0;
            for (Node n: nodes) {
                n.reset();
            }
            for (Arc a: arcs) {   // clear running/signalling arcs
                a.reset();        // from the previous runh
            }
            startNode.fire();

            simulationCycle:
            while (!que.isEmpty() && time <= stopAtTime) {
                event = que.remove();
                time = event.time;
                if (time > stopAtTime) { break simulationCycle; }
                event.arc.stop();
                if (!postponedEorNodes.isEmpty() && 
                        (que.isEmpty() || que.peek().time > time+delta)) {
                    for (Node n: postponedEorNodes) {
                        n.release();
                    }
                    postponedEorNodes.clear();
                }
            }
            if (time > stopAtTime) {
                log(1, time+":  End of simulation run "+run+
                        " -- final time "+stopAtTime+" exceded");
            }
            else if (que.isEmpty()) {
                log(1, time+":  Deadlock (empty queue) occured");
            }
        }
    }
    
    public void registerPostponedEorNode(Node n) {
        if (postponedEorNodes.isEmpty()) {
            postponedEorTime = time;
        }
        postponedEorNodes.add(n);
    }

    public void logc(int level, String txt) {
        if (level <= logLevel) {
            System.out.print(txt);
        }
    }

    public void log(int level, String txt) {
        if (level <= logLevel) {
            logc(level, txt + "\n");
        }
    }

    public void printStatistics() {
        log(0, "");
        log(0, "----- Statistics of "+nRuns+" simulation runs -----");
        log(0, "used random seed  "+usedRandSeed);
        log(0, "");

        for (Node n: nodes) {
            log(0, "Node  "+n.name+"\n"+n.stat.strStat());
        }
        log(0, "----- End of Gsi -----");
    }

}
